#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `pyn5` package."""


import unittest
from pathlib import Path
import shutil
import numpy as np


import pyn5


class BaseTestCase:
    class BaseTest(unittest.TestCase):
        def setUp(self):
            self.root = "test.n5"
            self.dataset = "test_{}".format(self.dtype)
            self.dataset_size = [10, 10, 10]
            self.block_size = [2, 2, 2]
            if Path(self.root).is_dir():
                shutil.rmtree(str(Path(self.root).absolute()))
            pyn5.create_dataset(
                self.root, self.dataset, self.dataset_size, self.block_size, self.dtype
            )
            self.n5 = pyn5.open(self.root, self.dataset, self.dtype, False)

        def tearDown(self):
            if Path(self.root).is_dir():
                shutil.rmtree(str(Path(self.root).absolute()))

        def test_read_write_valid(self):
            self.n5.write_block([0, 0, 0], self.valid_block)
            self.assertEqual(
                self.n5.read_ndarray([0, 0, 0], self.block_size), self.valid_block
            )

        def test_read_write_overflow(self):
            """
            Doesn't work properly for float types since floats that are too
            large get converted to inf
            """
            try:
                self.n5.write_block([1, 1, 1], self.overflow_block)
                raise AssertionError("Expected OverflowError")
            except OverflowError:
                self.assertEqual(
                    self.n5.read_ndarray([2, 2, 2], self.block_size),
                    [0] * np.prod(self.block_size),
                )

        def test_read_write_wrong_dtype(self):
            """
            Doesn't work properly for float types since I'm guessing ints
            get converted to float types somewhere between python and rust
            """
            try:
                self.n5.write_block([2, 2, 2], self.wrong_dtype_block)
                raise AssertionError("Expected TypeError")
            except TypeError:
                self.assertEqual(
                    self.n5.read_ndarray([4, 4, 4], self.block_size),
                    [0] * np.prod(self.block_size),
                )


class TestU8(BaseTestCase.BaseTest):
    def setUp(self):
        self.dtype = "UINT8"

        big = 2 ** 8
        self.valid_block = [0, 1, 2, 3, big - 4, big - 3, big - 2, big - 1]
        self.overflow_block = [-1, -2, -3, -4, big, big + 1, big + 2, big + 3]
        self.wrong_dtype_block = [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8]

        super().setUp()

    @unittest.expectedFailure
    def test_writting_wrong_dtype(self):
        bad_n5 = pyn5.open(self.root, self.dataset, "FLOAT64")
        try:
            bad_n5.write_block([0, 0, 0], [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8])
            self.fail("Expected TypeError")
        except TypeError:
            pass


class TestU16(BaseTestCase.BaseTest):
    def setUp(self):
        self.dtype = "UINT16"

        big = 2 ** 16
        self.valid_block = [0, 1, 2, 3, big - 4, big - 3, big - 2, big - 1]
        self.overflow_block = [-1, -2, -3, -4, big, big + 1, big + 2, big + 3]
        self.wrong_dtype_block = [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8]

        super().setUp()


class TestU32(BaseTestCase.BaseTest):
    def setUp(self):
        self.dtype = "UINT32"

        big = 2 ** 32
        self.valid_block = [0, 1, 2, 3, big - 4, big - 3, big - 2, big - 1]
        self.overflow_block = [-1, -2, -3, -4, big, big + 1, big + 2, big + 3]
        self.wrong_dtype_block = [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8]

        super().setUp()


class TestU64(BaseTestCase.BaseTest):
    def setUp(self):
        self.dtype = "UINT64"

        big = 2 ** 64
        self.valid_block = [0, 1, 2, 3, big - 4, big - 3, big - 2, big - 1]
        self.overflow_block = [-1, -2, -3, -4, big, big + 1, big + 2, big + 3]
        self.wrong_dtype_block = [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8]

        super().setUp()


class TestI8(BaseTestCase.BaseTest):
    def setUp(self):
        self.dtype = "INT8"

        big = 2 ** 7
        self.valid_block = [
            1 - big,
            2 - big,
            3 - big,
            4 - big,
            big - 4,
            big - 3,
            big - 2,
            big - 1,
        ]
        self.overflow_block = [
            -big,
            -1 - big,
            -2 - big,
            -3 - big,
            big,
            big + 1,
            big + 2,
            big + 3,
        ]
        self.wrong_dtype_block = [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8]

        super().setUp()


class TestI16(BaseTestCase.BaseTest):
    def setUp(self):
        self.dtype = "INT16"

        big = 2 ** 15
        self.valid_block = [
            1 - big,
            2 - big,
            3 - big,
            4 - big,
            big - 4,
            big - 3,
            big - 2,
            big - 1,
        ]
        self.overflow_block = [
            -big,
            -1 - big,
            -2 - big,
            -3 - big,
            big,
            big + 1,
            big + 2,
            big + 3,
        ]
        self.wrong_dtype_block = [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8]

        super().setUp()


class TestI32(BaseTestCase.BaseTest):
    def setUp(self):
        self.dtype = "INT32"

        big = 2 ** 31
        self.valid_block = [
            1 - big,
            2 - big,
            3 - big,
            4 - big,
            big - 4,
            big - 3,
            big - 2,
            big - 1,
        ]
        self.overflow_block = [
            -big,
            -1 - big,
            -2 - big,
            -3 - big,
            big,
            big + 1,
            big + 2,
            big + 3,
        ]
        self.wrong_dtype_block = [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8]

        super().setUp()


class TestI64(BaseTestCase.BaseTest):
    def setUp(self):
        self.dtype = "INT64"

        big = 2 ** 63
        self.valid_block = [
            1 - big,
            2 - big,
            3 - big,
            4 - big,
            big - 4,
            big - 3,
            big - 2,
            big - 1,
        ]
        self.overflow_block = [
            -big,
            -1 - big,
            -2 - big,
            -3 - big,
            big,
            big + 1,
            big + 2,
            big + 3,
        ]
        self.wrong_dtype_block = [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8]

        super().setUp()


class TestF32(BaseTestCase.BaseTest):
    def setUp(self):
        self.dtype = "FLOAT32"

        self.valid_block = [
            -3.3999999521443642e38,
            -1.199999978106707e-38,
            1.199999978106707e-38,
            3.3999999521443642e38,
            0.0,
            0.0,
            0.0,
            0.0,
        ]

        big = 1.5e40
        self.overflow_block = [
            -big,
            -1 - big,
            -2 - big,
            -3 - big,
            big,
            big + 1,
            big + 2,
            big + 3,
        ]
        self.wrong_dtype_block = [1, 2, 3, 4, 5, 6, 7, 8]

        super().setUp()

    @unittest.expectedFailure
    def test_read_write_overflow(self):
        super().test_read_write_overflow(self)

    @unittest.expectedFailure
    def test_read_write_wrong_dtype(self):
        super().test_read_write_wrong_dtype(self)


class TestF64(BaseTestCase.BaseTest):
    def setUp(self):
        self.dtype = "FLOAT64"

        self.valid_block = [-2.3e-308, -1.7e308, 1.7e308, 2.3e-308, 0.0, 0.0, 0.0, 0.0]

        big = 1.5e600
        self.overflow_block = [
            -big,
            -1 - big,
            -2 - big,
            -3 - big,
            big,
            big + 1,
            big + 2,
            big + 3,
        ]
        self.wrong_dtype_block = [1, 2, 3, 4, 5, 6, 7, 8]

        super().setUp()

    @unittest.expectedFailure
    def test_read_write_overflow(self):
        super().test_read_write_overflow(self)

    @unittest.expectedFailure
    def test_read_write_wrong_dtype(self):
        super().test_read_write_wrong_dtype(self)


class TestPythonReadWrite(unittest.TestCase):
    def setUp(self):
        self.root = "test.n5"
        self.dataset = "test"
        self.dtype = "UINT8"
        self.dataset_size = [10, 10, 10]
        self.block_size = [2, 2, 2]

        pyn5.create_dataset(
            self.root, self.dataset, self.dataset_size, self.block_size, self.dtype
        )
        self.n5 = pyn5.open(self.root, self.dataset, self.dtype, False)

    def tearDown(self):
        if Path(self.root).is_dir():
            shutil.rmtree(str(Path(self.root).absolute()))

    def test_read_write(self):
        # make sure n5 is initialized to all zeros
        self.assertTrue(
            np.array_equal(
                pyn5.read(self.n5, (np.array([0, 0, 0]), np.array(self.dataset_size))),
                np.zeros([10, 10, 10]),
            )
        )

        # write ones to whole dataset, and then
        # write on partial blocks to make sure data isn't overwritten
        pyn5.write(
            self.n5,
            (np.array([0, 0, 0]), np.array(self.dataset_size)),
            np.ones(self.dataset_size),
        )
        self.assertTrue(
            np.array_equal(
                pyn5.read(self.n5, (np.array([0, 0, 0]), np.array(self.dataset_size))),
                np.ones([10, 10, 10]),
            )
        )

        pyn5.write(
            self.n5, (np.array([1, 1, 1]), np.array([3, 3, 3])), np.ones([2, 2, 2]) * 2
        )
        self.assertTrue(
            np.array_equal(
                pyn5.read(self.n5, (np.array([1, 1, 1]), np.array([3, 3, 3]))),
                np.ones([2, 2, 2]) * 2,
            )
        )

        # test writting non-uniform block to make sure axis orderings are correct
        pyn5.write(
            self.n5,
            (np.array([1, 1, 1]), np.array([5, 5, 5])),
            np.array(range(64), dtype=int).reshape([4, 4, 4]),
        )
        self.assertTrue(
            np.array_equal(
                pyn5.read(self.n5, (np.array([1, 1, 1]), np.array([5, 5, 5]))),
                np.array(range(64), dtype=int).reshape([4, 4, 4]),
            )
        )




# -*- coding: utf-8 -*-

"""Main module."""
import json
from pathlib import Path
import logging
from typing import Tuple
import numpy as np

from .pyn5 import (
    DatasetUINT8,
    DatasetUINT16,
    DatasetUINT32,
    DatasetUINT64,
    DatasetINT8,
    DatasetINT16,
    DatasetINT32,
    DatasetINT64,
    DatasetFLOAT32,
    DatasetFLOAT64,
    create_dataset,
)


def open(root_path: str, dataset: str, dtype: str = "", read_only=True):
    """
    Returns a Dataset of the corresponding dtype. Leave dtype blank to return
    the Dataset with dtype as shown in the attributes.json file
    """

    # Check the attributes file:
    attributes_file = Path(root_path, dataset, "attributes.json")
    if attributes_file.exists():
        with attributes_file.open("r") as f:
            attributes = json.load(f)
        expected_dtype = attributes.get("dataType", None)
        if expected_dtype is not None:
            if dtype == "":
                # Use the expected dtype
                return open(root_path, dataset, expected_dtype.upper(), read_only)
            elif dtype != expected_dtype.upper():
                # When in doubt use the user specified dtype
                logging.warning(
                    "Given dtype {} does not match dtype ({}) in attributes.json".format(
                        dtype, expected_dtype.upper()
                    )
                )

    if dtype == "UINT8":
        return DatasetUINT8(root_path, dataset, read_only)
    elif dtype == "UINT16":
        return DatasetUINT16(root_path, dataset, read_only)
    elif dtype == "UINT32":
        return DatasetUINT32(root_path, dataset, read_only)
    elif dtype == "UINT64":
        return DatasetUINT64(root_path, dataset, read_only)
    elif dtype == "INT8":
        return DatasetINT8(root_path, dataset, read_only)
    elif dtype == "INT16":
        return DatasetINT16(root_path, dataset, read_only)
    elif dtype == "INT32":
        return DatasetINT32(root_path, dataset, read_only)
    elif dtype == "INT64":
        return DatasetINT64(root_path, dataset, read_only)
    elif dtype == "FLOAT32":
        return DatasetFLOAT32(root_path, dataset, read_only)
    elif dtype == "FLOAT64":
        return DatasetFLOAT64(root_path, dataset, read_only)
    else:
        raise ValueError(
            "Given dtype {} is not supported. Please choose from ({})".format(
                dtype,
                (
                    "UINT8",
                    "UINT16",
                    "UINT32",
                    "UINT64",
                    "INT8",
                    "INT16",
                    "INT32",
                    "INT64",
                    "FLOAT32",
                    "FLOAT64",
                ),
            )
        )


def read(dataset, bounds: Tuple[np.ndarray, np.ndarray], dtype: type = int):
    """
    Temporary hacky method until dataset.read_ndarray returns np.ndarray

    Note: passing in dtype is necessary since numpy arrays are float by default.
          dataset.get_data_type() could be implemented, but a better solution would
          be to have dataset.read_ndarray return a numpy array. 
    """
    bounds = (bounds[0].astype(int), bounds[1].astype(int))
    return (
        np.array(dataset.read_ndarray(list(bounds[0]), list(bounds[1] - bounds[0])))
        .reshape(list(bounds[1] - bounds[0]))
        .transpose([2, 1, 0])
        .astype(dtype)
    )


def write(
    dataset,
    input_bounds: Tuple[np.ndarray, np.ndarray],
    input_data: np.ndarray,
    dtype=int,
):
    """
    Temporary hacky method until dataset.write_ndarray is implemented in rust-n5
    and the PyO3 wrapper
    """
    input_data = input_data.astype(dtype)
    input_bounds = (input_bounds[0].astype(int), input_bounds[1].astype(int))
    start_block_index = input_bounds[0] // dataset.block_shape
    stop_block_index = (
        input_bounds[1] + dataset.block_shape - 1
    ) // dataset.block_shape
    for i in range(start_block_index[0], stop_block_index[0]):
        for j in range(start_block_index[1], stop_block_index[1]):
            for k in range(start_block_index[2], stop_block_index[2]):
                block_index = np.array([i, j, k], dtype=int)
                block_bounds = (
                    block_index * dataset.block_shape,
                    (block_index + 1) * dataset.block_shape,
                )

                if all(block_bounds[0] >= input_bounds[0]) and all(
                    block_bounds[1] <= input_bounds[1]
                ):
                    # Overwrite block data entirely
                    block_data = input_data[
                        tuple(
                            map(
                                slice,
                                block_bounds[0] - input_bounds[0],
                                block_bounds[1] - input_bounds[0],
                            )
                        )
                    ]
                else:
                    block_data = read(dataset, block_bounds, dtype)

                    intersection_bounds = (
                        np.maximum(block_bounds[0], input_bounds[0]),
                        np.minimum(block_bounds[1], input_bounds[1]),
                    )
                    relative_block_bounds = tuple(
                        map(
                            slice,
                            intersection_bounds[0] - block_bounds[0],
                            intersection_bounds[1] - block_bounds[0],
                        )
                    )
                    relative_data_bounds = tuple(
                        map(
                            slice,
                            intersection_bounds[0] - input_bounds[0],
                            intersection_bounds[1] - input_bounds[0],
                        )
                    )
                    block_data[relative_block_bounds] = input_data[relative_data_bounds]

                dataset.write_block(
                    block_index, block_data.transpose([2, 1, 0]).flatten()
                )
